local class = require 'class'

local lg = love.graphics
local gline, grect, gprint, gprintf, ggetColor, gsetColor, gnewFont, gsetFont = lg.line, lg.rectangle, lg.print, lg.printf, lg.getColor, lg.setColor, lg.newFont, lg.setFont
local kkeyIsDown = love.keyboard.isDown
local floor = math.floor

local MARK_SYMBOL, MINE_SYMBOL = '#', '@'
local CELL_FREE, CELL_MINE = 0, -1

local LoveSweeper = class()

LoveSweeper.init = function(self, kwargs)
	assert(kwargs)
	self.ui = kwargs.ui
	self.ui.font = gnewFont('04B_03__.TTF', kwargs.ui.fontsize)
	gsetFont(self.ui.font)

	self.keycounter = 0
	self.keydelay = kwargs.keydelay
	self.tapcounter = 0
	self.tapdelay = kwargs.tapdelay
	self.gamekwargs = kwargs.gamekwargs
	self.colors = kwargs.colors
	self.controls = kwargs.controls

	self.menu = {
		selected = 5,
		items = {
			{1, type = 'button', text = 'Novice',   rect = {10, 10, 90, 24}, move = {down = 2, up = 4, right = 7, left = 8}, onClick = function(self) self.gamekwargs.nmines = 10; self.gamekwargs.width = 9; self.gamekwargs.height = 9 end},
			{2, type = 'button', text = 'Advanced', rect = {10, 40, 90, 24}, move = {down = 3, up = 1, right = 9, left = 10}, onClick = function(self) self.gamekwargs.nmines = 40; self.gamekwargs.width = 16; self.gamekwargs.height = 16 end},
			{3, type = 'button', text = 'Expert',   rect = {10, 70, 90, 24}, move = {down = 4, up = 2, right = 11, left = 12}, onClick = function(self) self.gamekwargs.nmines = 99; self.gamekwargs.width = 30; self.gamekwargs.height = 16 end},

			{4, type = 'button', text = 'New',      rect = {10, 130, 90, 24}, move = {down = 1, up = 3, right = 5},onClick = function(self) self:restart(); self.showmenu = false end},
			{5, type = 'button', text = 'Cancel',   rect = {110, 130, 90, 24}, move = {down = 1, up = 3, left = 4, right = 6},onClick = function(self) self.showmenu = false end},
			{6, type = 'button', text = 'Exit',     rect = {210, 130, 90, 24}, move = {down = 1, up = 3, left = 5},onClick = function(self) self:quit() end},

			{7, type = 'button', text = '+',        rect = {210, 10, 24, 24}, move = {down = 9, up = 5, left = 1, right = 8}, onClick = function(self) if self.gamekwargs.nmines + 1 < self.gamekwargs.width * self.gamekwargs.height then self.gamekwargs.nmines = self.gamekwargs.nmines + 1 end end},
			{8, type = 'button', text = '-',        rect = {240, 10, 24, 24}, move = {down = 10, up = 5, left = 7, right = 1}, onClick = function(self) if self.gamekwargs.nmines - 1 > 1 then self.gamekwargs.nmines = self.gamekwargs.nmines - 1 end end},
			{9, type = 'button', text = '+',        rect = {210, 40, 24, 24}, move = {down = 11, up = 7, left = 2, right = 10}, onClick = function(self) if (self.gamekwargs.width+1)*self.ui.cellw <= self.ui.screenw then self.gamekwargs.width = self.gamekwargs.width + 1 end end},
			{10, type = 'button', text = '-',        rect = {240, 40, 24, 24}, move = {down = 12, up = 8, left = 9, right = 2}, onClick = function(self) if self.nmines < (self.gamekwargs.width-1) * self.gamekwargs.height then self.gamekwargs.width = self.gamekwargs.width - 1 end end},
			{11, type = 'button', text = '+',        rect = {210, 70, 24, 24}, move = {down = 5, up = 9, left = 3, right = 12}, onClick = function(self) if (self.gamekwargs.height+1)*self.ui.cellh <= self.ui.screenh then self.gamekwargs.height = self.gamekwargs.height + 1 end end},
			{12, type = 'button', text = '-',        rect = {240, 70, 24, 24}, move = {down = 5, up = 10, left = 11, right = 3}, onClick = function(self) if self.nmines < (self.gamekwargs.height-1) * self.gamekwargs.width then self.gamekwargs.height = self.gamekwargs.height - 1 end end},

			{type = 'text', text = 'mines %s', robj = self.gamekwargs, rfield = 'nmines', rect = {110, 10, -1, -1}},
			{type = 'text', text = 'width %s', robj = self.gamekwargs, rfield = 'width', rect = {110, 40, -1, -1}},
			{type = 'text', text = 'height %s', robj = self.gamekwargs, rfield = 'height', rect = {110, 70, -1, -1}},
			{type = 'text', text = 'LoveSweeper\nMade by Abraham Greed', rect = {10, 190, -1, -1}},
		}
	}

	self:restart(kwargs)
end

LoveSweeper.restart = function(self)
	local kwargs = self.gamekwargs

	assert(kwargs)
	self.width, self.height = kwargs.width, kwargs.height
	self.nmines = kwargs.nmines
	assert(self.nmines < self.width * self.height)
	self.nchecked = 0
	self.counter = 0
	self.gamestarted = false
	self.gamefinished = false
	self.selected = nil
	self.minemark = nil

	-- Set game field position
	self.ui.fieldx = floor(.5 * (self.ui.screenw - self.ui.cellw * kwargs.width) )
	self.ui.fieldy = floor(.5 * (self.ui.screenh - self.ui.cellh * kwargs.height) )

	-- Create field for mines
	self.field = {}
	self.checked = {}
	self.marked = {}
	self.selected = nil
	for x = 1, self.width do
		self.field[x] = {}
		self.checked[x] = {}
		self.marked[x] = {}
		for y = 1, self.height do
			self.field[x][y] = CELL_FREE
			self.checked[x][y] = false
			self.marked[x][y] = false
		end
	end
end

LoveSweeper.start = function(self, firstx, firsty)
	-- Fill field with mines
	local ncurrent = 0
	while ncurrent < self.nmines do
		local x, y = math.random(1, self.width), math.random(1, self.height)
		if not (x == firstx and y == firsty) and self.field[x][y] == 0 then
			ncurrent = ncurrent + 1
			self.field[x][y] = CELL_MINE
		end
	end

	-- Fill field with numbers of nearby mines count
	for x = 1, self.width do
		for y = 1, self.height do
			if self.field[x][y] ~= CELL_MINE then
				local n = 0
				if x > 1 then
					if self.field[x-1][y] == CELL_MINE then n = n + 1 end
					if y > 1 and self.field[x-1][y-1] == CELL_MINE then n = n + 1 end
					if y < self.height and self.field[x-1][y+1] == CELL_MINE then n = n + 1 end
				end

				if x < self.width then
					if self.field[x+1][y] == CELL_MINE then n = n + 1 end
					if y > 1 and self.field[x+1][y-1] == CELL_MINE then n = n + 1 end
					if y < self.height and self.field[x+1][y+1] == CELL_MINE then n = n + 1 end
				end

				if y > 1 and self.field[x][y-1] == CELL_MINE then n = n + 1 end
				if y < self.height  and self.field[x][y+1] == CELL_MINE then n = n + 1 end

				self.field[x][y] = n
			end
		end
	end

	self.gamestarted = true
end

LoveSweeper.press = function(self, key)
	self.keypress = key
	self.keycounter = 0
	self.tapcounter = 0
end

LoveSweeper.update = function(self, dt)
	local kp = self.keypress

	local isDirectionKey = function(key)
		local key = self.controls[key]
		return  key == 'left' or key == 'right' or key == 'down' or key == 'up'
	end

--~ 	if kp and self.selected or self.men
--~ 	isDirectionKey(kp) then
	if kp then
		if self.tapcounter >= self.tapdelay then
			self.keycounter = self.keycounter + dt
			if self.keycounter >= self.keydelay then
				self.keycounter = 0
				if kkeyIsDown(kp) then
					self:control(kp)
				end
			end
		else
			self.tapcounter = self.tapcounter + dt
		end
	end

	if self.gamestarted and not self.gamefinished and not self.showmenu then
		self.counter = self.counter + dt
	end
end


LoveSweeper.draw = function(self)
	local cw, ch = self.ui.cellw, self.ui.cellh
	local w, h = self.width, self.height
	local dh = .5 * (ch - self.ui.fontsize)
	local fx, fy = self.ui.fieldx, self.ui.fieldy

	gsetColor(self.colors.text)
	gprint('@ ' .. tostring(self.nmines), self.ui.fieldx, self.ui.fieldy - self.ui.fontsize - 4)

	local t = floor(self.counter)
	local hs = floor(t / 3600)
	local ms = floor((t - hs * 3600) / 60)
	local ss = t - hs * 3600 - ms * 60
	gprint(('%8d:%02d:%02d'):format(hs, ms, ss), self.ui.fieldx + self.ui.cellw * self.width - 7*self.ui.fontsize, self.ui.fieldy - self.ui.fontsize - 4)

	gsetColor(self.colors.checked)
	grect('fill', fx, fy, w * cw, h * ch)

	for x = 1, self.width do
		for y = 1, self.height do
			if self.checked[x][y] then
				if self.field[x][y] ~= CELL_FREE and self.field[x][y] ~= CELL_MINE then
					local num = self.field[x][y]
					gsetColor(self.colors.numbers[num])
					gprintf(tostring(num), fx + (x-1) * cw, dh + fy + (y-1) * ch, cw, 'center')
				elseif self.field[x][y] == CELL_MINE then
					gsetColor(self.colors.mine)
					gprintf(MINE_SYMBOL, fx + (x-1) * cw, dh + fy + (y-1) * ch, cw, 'center')
				end
			else
				gsetColor(self.colors.notchecked)
				grect('fill', fx + (x-1) * cw, fy + (y-1) * ch, cw, ch)
			end

			if self.marked[x][y] then
				gsetColor(self.colors.marked)
				gprintf(MARK_SYMBOL, fx + (x-1) * cw, dh + fy + (y-1) * ch, cw, 'center')
			end

		end
	end

	if self.minemark then
		local x, y = self.minemark[1], self.minemark[2]
		gsetColor(self.colors.exploded)
		grect('fill', fx + (x-1) * cw, fy + (y-1) * ch, cw, ch)
	end

	if self.selected then
		local x, y = self.selected[1], self.selected[2]
		gsetColor(self.colors.selected)
		grect('fill', fx + (x-1) * cw, fy + (y-1) * ch, cw, ch)
	end

	gsetColor(self.colors.lines)
	for x = 0, self.width do
		gline(fx + x * cw, fy, fx + x * cw, fy + h * ch)
	end

	for y = 0, self.height do
		gline(fx, fy + y * ch, fx + w * cw, fy + y * ch)
	end

	if self.showmenu then
		local mw, mh = self.ui.menuw, self.ui.menuh
		local mx, my = self.ui.menux, self.ui.menuy
		gsetColor(self.colors.menubackground)
		grect('fill', mx, my, mw, mh)

		for i = 1, #self.menu.items do
			local item = self.menu.items[i]
			if item.type == 'button' then
				gsetColor(self.colors.buttoncolor)
				grect('fill', mx + item.rect[1], my + item.rect[2], item.rect[3], item.rect[4])
				gsetColor(self.colors.buttontextcolor)
				gprintf(item.text, mx + item.rect[1], 4 + my + item.rect[2], item.rect[3], 'center')
			elseif item.type == 'text' then
				gsetColor(self.colors.textcolor)
				local text = item.text
				if item.robj and item.rfield then
					text = text:format(tostring(item.robj[item.rfield]))
				end
				gprint(text, mx + item.rect[1], 4 + my + item.rect[2])
			end
		end

		local item = self.menu.items[self.menu.selected]
		gsetColor(self.colors.buttonselectedcolor)
		grect('line', mx + item.rect[1], my + item.rect[2], item.rect[3], item.rect[4])

	end
end

local MOUSE_ACTIONS = {l = 'check', r = 'mark'}

LoveSweeper.clicked = function(self, pointerx, pointery, button)
	if not self.showmenu then
		local fx, fy = self.ui.fieldx, self.ui.fieldy
		local cw, ch = self.ui.cellw, self.ui.cellh
		local w, h = self.width, self.height
		local ceil = math.ceil

		if pointerx > fx and pointery > fy and pointerx < fx + w * cw and pointery < fy + h * ch then
			local x, y = ceil((pointerx - fx) / cw), ceil((pointery - fy) / ch)
			self:clickCell(x, y, MOUSE_ACTIONS[button])
		else
			self.showmenu = true
		end
	else
		local x, y = pointerx, pointery
		for i = 1, #self.menu.items do
			local item = self.menu.items[i]
			if item.type == 'button' then
				local r = item.rect
				if x >= r[1] and x <= r[1] + r[3] and y >= r[2] and y <= r[2] + r[4] then
					self.menu.selected = i
					item.onClick(self)
					break
				end
			end
		end
	end
end

LoveSweeper.clickCell = function(self, x, y, action)
	if action == 'check' or action == 'mark' then
		if self.gamefinished then
			self:restart()
			return
		end

		if not self.gamestarted then
			self:start(x, y)
		end

		if not self.checked[x][y] then
			if action == 'mark' then
				self.marked[x][y] = not self.marked[x][y]
			elseif not self.marked[x][y] then
				self:check(x, y)
			end
		end
	end
end

LoveSweeper.check = function(self, x, y)
	self.nchecked = self.nchecked + 1
	self.checked[x][y] = true
	self.marked[x][y] = false

	if self.nchecked >= self.width * self.height - self.nmines then
		self:win()
	end

	if self.field[x][y] == CELL_MINE then
		self:gameover(x, y)
	elseif self.field[x][y] == CELL_FREE then
		if x > 1 and not self.checked[x-1][y] then
			self:check(x-1, y)
		end

		if x < self.width and not self.checked[x+1][y] then
			self:check(x+1, y)
		end

		if y > 1 and not self.checked[x][y-1] then
			self:check(x, y-1)
		end

		if y < self.height and not self.checked[x][y+1] then
			self:check(x, y+1)
		end
	end
end

LoveSweeper.gameover = function(self, x, y)
	self.minemark = {x, y}
	for x = 1, self.width do
		for y = 1, self.height do
			self.marked[x][y] = false
			self.checked[x][y] = true
		end
	end

	self.gamefinished = true
end

LoveSweeper.win = function(self)
	for x = 1, self.width do
		for y = 1, self.height do
			self.marked[x][y] = false
			self.checked[x][y] = true
		end
	end

	self.selected = nil
	self.gamefinished = true
end

LoveSweeper.control = function(self, key)
	local key = self.controls[key]
	if key == 'quit' then
		self.showmenu = not self.showmenu
	end

	if not self.showmenu then
		if not self.selected then
			self.selected = {math.ceil(.5 * self.width), math.ceil(.5 * self.height)}
		end

		if key == 'left' and self.selected[1] > 1 then self.selected[1] = self.selected[1] - 1 end
		if key == 'right' and self.selected[1] < self.width then self.selected[1] = self.selected[1] + 1 end
		if key == 'up' and self.selected[2] > 1 then self.selected[2] = self.selected[2] - 1 end
		if key == 'down' and self.selected[2] < self.height then self.selected[2] = self.selected[2] + 1 end

		if key == 'check' or key == 'mark' then
			self:clickCell(self.selected[1], self.selected[2], key)
		end
	else
		self:controlMenu(key)
	end
end

LoveSweeper.controlMenu = function(self, key)
	if key == 'left' or key == 'right' or key == 'down' or key == 'up' then
		if self.menu.items[self.menu.selected].move[key] then
			self.menu.selected = self.menu.items[self.menu.selected].move[key]
		end
	elseif key == 'check' then
		self:clickButton()
	end
end

LoveSweeper.clickButton = function(self)
	if self.menu.selected then
		self.menu.items[self.menu.selected].onClick(self)
	end
end

LoveSweeper.quit = function(self)
	if love.event.quit then
		love.event.quit()
	else
		love.event.push('q')
	end
end

return LoveSweeper
